/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2015 Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-system.c 126 2013-04-11 08:39:36Z acau $
 */


#include "system.h"





/************************\
*                        *
* top-level run command  *
*                        *
\************************/

/* run a tempura program. */
void reduce_RUN (e)
NODEPTR e;
{
    NODEPTR emptyflag;
    unsigned old_run_flag = RunFlag;
    NODE local_empty[MEMCELLSIZE];
    struct timeval tvBegin, tvEnd, tvDiff;
    
    push_empty (new_loc2 (local_empty, STATE, 1, 0));
    RunFlag = T;
    node_lnk_2(e) = node_lnk_1(e);
    
    gettimeofday(&tvBegin, NULL);
    init_exec_stat ();
    do {
        if (TimeStep > 0)
            start_time ();
        init_state_stat ();
        do {
            init_pass_stat ();
            node_lnk_2(e) = reduce_pgm (node_lnk_2(e));
            output_pass_stat ();
            update_state_stat ();
        } while (!Ready);

        output_state_stat ();
        update_exec_stat ();
        emptyflag = get_empty ();
        if (emptyflag == False) {
            StateNumber += 1;
            Clock += 1;
            if (!OneLevelMemFlag) {
                switch_memory;
    	        EmptyLoc = curr_loc (EmptyLoc);
            }
	    if (TimeStep > 0)
	        check_time ();
        }
    } while (emptyflag != True);

    output_exec_stat ();
    gettimeofday(&tvEnd, NULL);
    timeval_subtract(&tvDiff, &tvEnd, &tvBegin);
#if defined(__APPLE__) && defined(__MACH__)    
    printf("Time elapsed: %ld.%06d\n", tvDiff.tv_sec, tvDiff.tv_usec);
#else
    printf("Time elapsed: %ld.%06ld\n", tvDiff.tv_sec, tvDiff.tv_usec);
#endif    
    collect_garbage();
    pop_empty;
    RunFlag = old_run_flag;
    result (node_lnk_2(e));
}

/* initialize statistics, etc. for execution */
void init_exec_stat ()
{
    TotalReduceCount = 0;
    MaxReduceDepth = 0;
    StateNumber = 0;
    TotalSuccessfulReduceCount = 0;
    TotalPassCount = 0;
}

/* initialize statistics, etc. for state */
void init_state_stat ()
{
    StateReduceCount = 0;
    StateSuccessfulReduceCount = 0;
    PassNumber = 0;
    PassLimitCounter = 0;
    StateProgressFlag = T;
    ForceOutput = F;
    
}

/* initialize statistics, etc. for state */
void init_pass_stat ()
{
    PassReduceCount = 0;
    PassSuccessfulReduceCount = 0;
    PassProgressFlag = F;
}

/* update execution statistics */
void update_exec_stat ()
{
   TotalReduceCount += StateReduceCount;
    TotalSuccessfulReduceCount += StateSuccessfulReduceCount;
    TotalPassCount += PassNumber;
}

/* update statistics, etc. for state */
void update_state_stat ()
{
    StateReduceCount += PassReduceCount;
    StateSuccessfulReduceCount += PassSuccessfulReduceCount;
    PassNumber += 1;
    PassLimitCounter += 1;
    StateProgressFlag = !ForceOutput;
    ForceOutput = !PassProgressFlag;
}

/* output statistics after each pass. */
void output_pass_stat ()
{
    if (state_debug) {
        output_state_number;
        output_str ("Program: ");
        output_expn (node_lnk_2(ReduceStackTop -> n), PrintLevel);
        newline;
    }
    if (pass_verbosity) {
        printf ("Pass #%d complete.  Reductions:  %d  (%d successful).\n",
                PassNumber, PassReduceCount, PassSuccessfulReduceCount);
        fflush(stdout);
    }
}

/* output out some statistics at the end of each state */
void output_state_stat ()
{
    if (state_verbosity) {
        printf ("State #%d ready.  Passes:  %d.  ", StateNumber, PassNumber);
        printf ("Reductions:  %d  (%d successful).\n",
                StateReduceCount, StateSuccessfulReduceCount);
        fflush(stdout);
        if (store_verbosity) {
            printf ("Memory in use:  ");
            printf ("Global:  %d ", GlobalMemUsed);
            printf ("(%d%%).  ",(GlobalMemUsed*100)/GLOBALMEMSIZE);
            printf ("Local:  %d.  ", LocalMemUsed);
            printf ("Heap:  %d ", HeapSize-FreeNodes);
            printf ("(%d%%).\n",((HeapSize-FreeNodes)*100)/HeapSize);
            fflush(stdout);
        }
        newline;
        fflush(stdout);
    }
    else 
	fputs ("", stdout);
}

void output_exec_stat ()
{
    if (exec_verbosity) {
        printf ("\nDone!  Computation length:  %d.  Total Passes:  %d.\n",
                StateNumber, TotalPassCount);
        fflush(stdout);
        printf ("Total reductions:  %d  (%d successful).  ",
                TotalReduceCount, TotalSuccessfulReduceCount);
        printf ("Maximum reduction depth:  %d.\n", MaxReduceDepth);
        fflush(stdout);
        if (store_verbosity) {
            printf ("Memory in use:  ");
            printf ("Global:  %d ", GlobalMemUsed);
            printf ("(%d%%).  ",(GlobalMemUsed*100)/GLOBALMEMSIZE);
            printf ("Local:  %d.  ", LocalMemUsed);
            printf ("Heap:  %d ", HeapSize-FreeNodes);
            printf ("(%d%%).\n",((HeapSize-FreeNodes)*100)/HeapSize);
            fflush(stdout);
        }
    }
}



/*************************\
*                         *
*  other system commands  *
*                         *
\*************************/

/* read in a file */
/* LOAD e */
void reduce_LOAD (e)
NODEPTR e;
{
        /*int     old_input_line_count = yylineno;*/
    char   *old_filename = InputFilename,
            name[MAXFILENAMESIZE];
    FILE * fp = NULL,
        *openfile ();
    NODEPTR e1 = reduce_val (node_lnk_1(e));

    if (Ready) {
        type_check (e1, STRING);
        /* Next line changed to avoid problems with garbage collector.
	   BCM  14Dec96 */
        if ((fp = openfile (name, node_strval(Path), node_strval(e1), FX, "r")) == NULL) {
        /* if ((fp = openfile (name, Path, e1 -> c1, FX, "r")) == NULL) {*/
            warn1 ("cannot open file \"%s\"",  node_strval(e1));
        }
        else {
            InputFilename = name;
            yylineno=1;Linetext[0]='\0';plinetext=Linetext;yycolnum = 1;
            printf ("[Reading file %s]\n", InputFilename);
            fflush(stdout);
            INTERACTIVE = 0;
            while (!feof (fp)) {
                exec_command (parse (fp));
            }
            fclose (fp);
            INTERACTIVE = 1;
        }
        result (True);
    }
    InputFilename = old_filename;
        /*yylineno = old_input_line_count;*/
}


/* define a value or function */
/* {define e1 e2} */
void reduce_DEFINE(e)
NODEPTR e;
{
    NODEPTR id = node_lnk_1(e);
    NODEPTR e2;
    assert1 (is_identifier (id), "cannot redefine operator: \"%s\"",
            print_name (id));
    e2 = reduce_val (node_lnk_2(e));
    if (Ready) {
        make_static_mem (mem_loc (id));
        put_mem1 (curr_loc (mem_loc (id)), e2);
        Ready = T;
        result (True);
    }
}

/* EXIT */
/* exit tempura to system. */
void reduce_EXIT(e)
NODEPTR e;
{
/*    intrrest (CNTRL_C);*/
    if (Fpcur != NULL)
        fclose(Fpcur);
    exit(0);
}

/* SHOWSTATE */
/* show the value (both defined and undefined) of the variables
   in the current state. */
void reduce_SHOWSTATE(e)
NODEPTR e;
{
    show_state();
    Ready = T;
    result (True);
    
}

/* SET sysid e2 */
void reduce_SET(e)
NODEPTR e;
{
    char   name[MAXFILENAMESIZE];
    FILE *openfile ();
    NODEPTR old_val, new_val;
    old_val = reduce_val (node_lnk_1(e));
    if (Ready) {
	new_val = reduce_val (node_lnk_2(e));
	if (Ready) {
	    type_check (new_val, node_type(old_val));
            printf ("%s was changed from ", print_name (node_lnk_1(e)));
            output_val (old_val);
            printf (" to ");
            output_val (new_val);
            newline;
            fflush(stdout);
	    assign_sysvar (node_lnk_1(e), new_val);
            if (node_type(node_lnk_1(e)) == INFILE) {
                if (strcmp (node_strval(OldInFile), "stdin") != 0) {
                    if (Fpcur != NULL) {
                        fclose (Fpcur);
                    }
                }
                if (strcmp (node_strval(InFile), "stdin") != 0) {
                    if ((Fpcur = openfile (name, node_strval(Path), node_strval(InFile), FXIN, "r")) == NULL) {
                        warn1 ("cannot open file \"%s\"",  node_strval(InFile));
                    }
                    else {
                     InFilename = name;
                         /*yylineno = 1;*/
                     printf ("[Input file is %s]\n", InFilename);
                     EOFILE=0;
                     fflush(stdout);
                    }
                }
            }
	    result (True);
	}
    }
}

void reduce_SYSVAR(e)
NODEPTR e;
{
    switch (node_type(e)) {
	case PASSLIMIT:
	    e = int_node (PassLimit);
	    break;
	case VERBOSITY:
	    e = int_node (Verbosity);
	    break;
	case PRINTLEVEL:
	    e = int_node (PrintLevel);
	    break;
	case DEBUGLEVEL:
	    e = int_node (DebugLevel);
	    break;
	case STATEPRINT:
	    e = bool_node (StatePrintFlag);
	    break;
	case REDUNDANCY:
	    e = bool_node (RedundancyFlag);
	    break;
	case CALLBYNAME:
	    e = bool_node (CallByNameFlag);
	    break;
	case ASSIGNAHEAD:
	    e = bool_node (AssignAheadFlag);
	    break;
        case BREAKISABORT:
	    e = bool_node (BreakIsAbortFlag);
	    break;    
        case HOPCHOP:
            e = bool_node (HopChop);
	    break;
	case PATH:
	    e = Path;/*  Avoid problems with garbage collector.
			    BCM  14Dec96 */
                /* e = string_node (Path);*/
	    break;
	case HELPDIR:
	    e = HelpDir;/*  Avoid problems with garbage collector.
			      BCM  14Dec96 */
                /* e = string_node (HelpDir);*/
	    break;
        case INFILE:
            e = InFile;
            break;
        case OUTFILE:
            e = OutFile;
            break;
	case TIMESTEP:
	    e = int_node (TimeStep);
	    break;
        case RANDSEED:
            e = int_node (RandSeed); /* AC 29Apr1999 */
            break;
        case RRANDSEED:
            e = int_node (RRandSeed); /* AC 20/09/2000 */
            break;
         case SRANDSEED:
            e = int_node (SRandSeed); 
            break;
        case SRRANDSEED:
            e = int_node (SRRandSeed); 
            break;   
        case MAXRANDLEN:
            e = int_node (MaxRandLen); /* AC 29Apr1999 */
            break;
        case PRECISION:
            e = int_node (Precision); /* AC 29Apr1999 */
            break;    
	default:
	    error_break1 ("not a system variable: %s", print_name (e));
    }
    Ready = T;
    result (e);
}


/******************\
*                  *
* system commands  *
*                  *
\******************/


/* SHOWTRACE */
void reduce_SHOWTRACE(e)
NODEPTR e;
{
    show_scope();
    show_trace();
    Ready = T;
    result (True);
}

/*
void show_mem(loc)
NODEPTR loc;
{
    
    printf (is_static_mem (loc) ? "*  " : "   ");
    printf ("(%4d;%4d;%2d)%s%s%s(%p)->%p:", Clock,node_mem_time(loc),node_mem_memno(loc),adj_loc (loc) == loc ? "1" : "0",mem_defined (loc)? "+" : "-",is_prevprev (loc) ? "PP" : is_prevnext (loc) ? "PN" : is_prev (loc) ? "P " : is_next (loc) ? "N " : "C ",loc,node_next(loc));
    while (is_indirect_mem (loc)) {
        loc = node_next(loc);
        printf ("%p%s -> ", cell_no (loc), is_prevprev (loc) ? "pp" : is_prevnext (loc) ? "pn" : is_prev (loc) ? "p " : is_next (loc) ? "n " : "c ");
    }
    if (mem_defined (loc))
        output_expn (node_next(loc), -1);
    else
    {
        output_expn (node_next(loc), -1);
    }
}
*/

void show_mem1(loc)
NODEPTR loc;
{
    unsigned int i;
    
    printf (is_static_mem (loc) ? "*  " : "   ");
    printf ("[%3d;%3d;%3d;%3d]%s%s%d", Clock,node_mem_time(loc),node_mem_memno(loc),node_mem_framed(loc),adj_loc1 (loc) == loc ? "1" : "0",mem_defined1 (loc)? "+" : "-",node_mem_history(loc));
    if (is_curr(loc)) {
        printf ("C ");
    }
    else {
        if (is_next(loc)) {
            printf ("N ");
        }
        else {
            if (is_prev1(loc,0)) {
                printf ("PC");
            }
            else {
                if (is_prev1(loc,1)) {
                    printf ("PN");
                }
                else {
                    for (i=2; i<=1+node_mem_history(loc); i=i+1)
                        if (is_prev1(loc,i))
                            printf ("%s%d","P",node_mem_history(loc)+2-i);
                }
            }
        }
    }
    
    printf ("(%p)->%p:",loc,node_next(loc));
    while (is_indirect_mem (loc)) {
        loc = node_next(loc);
        printf ("%p",cell_no (loc));
        if (is_curr(loc)) {
            printf ("c ");
        }
        else {
            if (is_next(loc)) {
                printf ("n ");
            }
            else {
                if (is_prev1(loc,0)) {
                    printf ("pc");
                }
                else {
                    if (is_prev1(loc,1)) {
                        printf ("pn");
                    }
                    else {
                        for (i=2; i<=1+node_mem_history(loc); i=i+1)
                            if (is_prev1(loc,i))
                                printf ("%s%d","p",node_mem_history(loc)+2-i);
                    }    
                }
            }            
        }
        printf(" -> ");
    }
    if (node_next(loc) != NULL){
        if (node_type(node_next(loc))==LIST) {
            if (is_curr(loc))
                output_list_curr(node_next(loc));
            if (is_next(loc))
                output_list_next(node_next(loc));
            for (i=0; i<=1+node_mem_history(loc); i=i+1)
                if (is_prev1(loc,i))
                    output_list_prev(node_next(loc),i);
        }
        else {
            output_expn (node_next(loc), -1);
        }
    }
    else {
        output_expn (node_next(loc), -1);
    }
    
        
    
    
}

/*
void show_mem_all(loc)
NODEPTR loc;
{
    NODEPTR cloc,nloc,ploc,pnloc,pploc;
    cloc = curr_loc(loc);
    nloc = next_loc(loc);
    ploc = prev_loc(loc);
    pnloc = prevnext_loc(loc);
    pploc = prevprev_loc(loc);

    printf ("NF=%1d, PF=%1d\n",NextFlag, PrevFlag);
    show_mem(cloc);
    newline;
    show_mem(nloc);
    newline;
    show_mem(ploc);
    newline;
    show_mem(pnloc);
    newline;
    show_mem(pploc);
    newline;
        
}
*/

void show_mem_all1(loc)
NODEPTR loc;
{
    unsigned int i;

        /*printf ("NF=%1d, PF=%1d\n",NextFlag, PrevFlag);*/
    show_mem1(curr_loc(loc));
    newline;
    show_mem1(next_loc(loc));
    newline;
    for (i=0; i<=1+node_mem_history(loc); i = i + 1) {
        show_mem1(prev_loc1(loc,i));
        newline;
    }
    
        
}

void show_mem_nodef(loc)
NODEPTR loc;
{
    unsigned int i;
    printf (is_static_mem (loc) ? "*  " : "   ");
    printf ("<%4d;%4d;%2d>  ", Clock,node_mem_time(loc),node_mem_memno(loc));
    while (is_indirect_mem (loc)) {
        loc = node_next(loc);
        printf ("%p",cell_no (loc));
        if (is_curr(loc)) {
            printf ("c ");
        }
        else {
            if (is_next(loc)) {
                printf ("n ");
            }
            else {
                if (is_prev1(loc,0)) {
                    printf ("pc");
                }
                else {
                    if (is_prev1(loc,1)) {
                        printf ("pn");
                    }
                    else {
                        for (i=2; i<=1+node_mem_history(loc); i=i+1)
                            if (is_prev1(loc,i))
                                printf ("%s%d","p",node_mem_history(loc)+2-i);
                    }    
                }
            }            
        }
        printf(" -> ");
    }
    output_expn (node_next(loc), -1);
}


/* SHOWMEM e1 e2 */
void reduce_SHOWMEM(e)
NODEPTR e;
{
    NODEPTR e1, e2;
    NODEPTR loc;
    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (e1, INT);
        e2 = reduce_val (node_lnk_2(e));
        if (Ready) {
            type_check (e2, INT);
            printf ("\nLocation  S  Time  Value\n");
            fflush(stdout);
            for (loc = node_lnk_1(e1); loc <= node_lnk_1(e2); loc++) {
                printf ("%7pc  ", loc);
                show_mem1 (curr_loc (loc));
                printf ("\n%8s  ", "n");
                show_mem1 (next_loc (loc));
                printf ("\n");
                fflush(stdout);
            }
            printf("\n");
            fflush(stdout);
            result (True);
        }
    }
}

/* EOF (NULL) */
void reduce_EOF(e)
NODEPTR e;
{
    if (InputFile == stdin)
        pop_tempura_level (T);
    Ready = T;
    result (True);
}

/* Pop up to next higher level. */
/* RETURN */
void reduce_RETURN(e)
NODEPTR e;
{
    pop_tempura_level (T);
    Ready = T;
    result (True);
}

/* Pop up to next higher level. */
/* CONTINUE */
void reduce_CONTINUE(e)
NODEPTR e;
{
    pop_tempura_level (F);
    Ready = T;
    result (True);
}

/* LEAP */
void reduce_LEAP(e)
NODEPTR e;
{
    BreakDepth = ReduceDepth - 1;/* Already reducing LEAP. */
    pop_tempura_level (F);
    Ready = T;
    result (True);
}

/* STEP */
void reduce_STEP(e)
NODEPTR e;
{
    BreakDepth = ReduceDepth;   /* Already reducing STEP. */
    pop_tempura_level (F);
    Ready = T;
    result (True);
}

/* Call top-level recursively. */
/* BREAK */
void reduce_BREAK(e)
NODEPTR e;
{
    user_break ("program break");
    Ready = T;
    result (True);
}

/* ABORT */
void reduce_ABORT(e)
NODEPTR e;
{
    TempuraLevel = 0;
    Break = T;
    Ready = T;
    result (True);
}

/* print a help message. */
void reduce_HELP(e)
NODEPTR e;
{
    char    ch,
           *subject = "summary",
            name[MAXFILENAMESIZE];
    FILE * fp, *openfile ();
    NODEPTR sp;

    Ready = T;
    if (node_lnk_1(e) != NULL) {
        sp = reduce_val (node_lnk_1(e));
        if (Ready && node_type(sp) == STRING)
            subject = node_strval(sp);
    }
    /* Next line changed to avoid problems with garbage collector.
      BCM  14Dec96 */
    if ((fp = openfile (name, node_strval(HelpDir), subject, "", "r")) == NULL) {
/*    if ((fp = openfile (name, HelpDir, subject, "", "r")) == NULL) {*/
        printf ("***No help available");
        if (node_lnk_1(e) == NULL)
            printf ("\n");
        else
            printf (" on \"%s\"\n", subject);
        if ((fp = openfile (name, node_strval(HelpDir), "summary", "", "r")) != NULL) {
        while ((ch = getc (fp)) != EOF)
            putchar (ch);
        fclose (fp);
        }
        fflush(stdout);
    }
    else {
        while ((ch = getc (fp)) != EOF)
            putchar (ch);
        fclose (fp);
    }
    result (True);
}

void reduce_SYSTEM(e)
NODEPTR e;
{
    NODEPTR e1 = reduce_val (node_lnk_1(e));
#ifdef HAVE_WORKING_FORK
    if (Ready) {
        int pid;
	type_check (e1, STRING);
	pid = fork ();
        if (pid == 0)
            exit (system (node_strval(e1)));
	put_empty (False);
	result (node1 (iSYSTEM, int_node (pid)));
    }
#endif
}

void reduce_iSYSTEM(e)
NODEPTR e;
{
    int pid = node_ival(node_lnk_1(e));
    NODEPTR *wl = &SysWaitList;
    check_sys_waits ();
    Ready = T;
    while (*wl != NULL) {
	if (node_ival(node_lnk_1((*wl)) ) == pid) {
	    *wl = node_lnk_2((*wl));
	    put_empty (True);
	    result (True);
	    return;
	}
	wl = &(node_lnk_2((*wl)));
    }
    put_empty (False);
}

	    
